-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_MPlayer"

-- **************************************************
-- General information about this script
-- **************************************************

FA_MPlayer = {}

function FA_MPlayer:Name()
	return "Player"
end

function FA_MPlayer:Version()
	return "5.0"
end

function FA_MPlayer:Description()
	return "MPlayer frontend to play external files"
end

function FA_MPlayer:LoadPrefs(prefs)
	self.MP_Invoke= prefs:GetString("FA_MPlayer.MP_Invoke", "")
	self.MP_ExeFormat= prefs:GetString("FA_MPlayer.MP_ExeFormat", "")
	self.MP_Fifo= prefs:GetString("FA_MPlayer.MP_Fifo", "")
	self.MP_ExeMode= prefs:GetBool("FA_MPlayer.MP_ExeMode", false)
end

function FA_MPlayer:SavePrefs(prefs)
	prefs:SetString("FA_MPlayer.MP_Invoke",self.MP_Invoke)
	prefs:SetString("FA_MPlayer.MP_ExeFormat",self.MP_ExeFormat)
	prefs:SetString("FA_MPlayer.MP_Fifo",self.MP_Fifo)
	prefs:SetBool("FA_MPlayer.MP_ExeMode",self.MP_ExeMode)
end

-- **************************************************
-- Recurring values
-- **************************************************

-- invoking MPlayer in slave mode

FA_MPlayer.MP_Windows_Invoke = "\"C:\\Program Files\\MPlayer\\mplayer.exe\" -slave -ontop -loop 0 -fixed-vo -geometry 50%:5%"
FA_MPlayer.MP_Linux_Invoke = "mplayer -slave -ontop -loop 0 -fixed-vo -geometry 50%:5%"
FA_MPlayer.MP_Mac_Invoke = "mplayer -slave -ontop -loop 0 -fixed-vo -geometry 50%:5%"

-- invoking with different file types (begins with separator space!)

FA_MPlayer.MP_PlayFile = " \"%s\""				--play a movie file (mpg, avi, mov etc.)
FA_MPlayer.MP_PlaySeq = " \"mf://%s\" -mf fps=%d:type=%s"	--play an image sequence (jpg, tga, png)
FA_MPlayer.MP_PlaySound = " -audiofile \"%s\""			--additional audio file (for image sequences)
FA_MPlayer.MP_PlaySub = " -sub \"%s\""				--additional subtitle file

-- configurable control commands for MPlayer

FA_MPlayer.MP_Quit = "quit\n"
FA_MPlayer.MP_Init = "pausing seek 0\nframe_step\n"	--after opening, pause on the first frame
FA_MPlayer.MP_Seek = "seek %d\n"				--fast seek
FA_MPlayer.MP_Seek_Pause = "pausing seek %d\n"		--fast seek and pause
FA_MPlayer.MP_Stop = "frame_step\n"			--stop (pause) playing
FA_MPlayer.MP_Start = "pause\n"				--start playing
FA_MPlayer.MP_Step = "frame_step\n"			--step one frame and pause
FA_MPlayer.MP_Step_Back = "pausing seek -2\n"		--step back and pause
FA_MPlayer.MP_OSD_On = "osd 2\n"				--switch on OSD displaying
FA_MPlayer.MP_OSD_On_Pause = "osd 2\npausing seek 0\n"	--switch off OSD displaying and pause
FA_MPlayer.MP_OSD_Off = "osd 0\n"			--switch off OSD displaying
FA_MPlayer.MP_OSD_Off_Pause = "pausing osd 0\n"		--switch off OSD displaying and pause

FA_MPlayer.MP_On = false
FA_MPlayer.MP_File = nil
FA_MPlayer.MP_OSDMode = false
FA_MPlayer.MP_OSD_Enabled = true
FA_MPlayer.MP_Playing = false

FA_MPlayer.MP_Invoke = ""
FA_MPlayer.MP_FullInvoke = ""
FA_MPlayer.MP_ExeFormat = ""
FA_MPlayer.MP_ExeMode = false

FA_MPlayer.playList= {}

-- **************************************************
-- The guts of this script
-- **************************************************

function FA_MPlayer:IsEnabled(moho)
	return true
end

-- **************************************************
-- play list dialog window
-- **************************************************

local FA_MPlayerListDialog = {}

function FA_MPlayerListDialog:new()
	local dlg = LM.GUI.SimpleDialog("Play List", FA_MPlayerListDialog)

	if (not dlg) then
		return nil
	end
	local l = dlg:GetLayout()

	l:PushH()
		l:PushV(LM.GUI.ALIGN_TOP)
			l:AddChild(LM.GUI.StaticText("Play list:"))
			dlg.list= LM.GUI.TextList(160,200,FA_MPlayer.LIST)
			l:AddChild(dlg.list)

			l:PushH(LM.GUI.ALIGN_CENTER)
				dlg.upButt= LM.GUI.Button(" < Up ",FA_MPlayer.UP)
				l:AddChild(dlg.upButt)
				dlg.downButt= LM.GUI.Button(" Down > ",FA_MPlayer.DOWN)
				l:AddChild(dlg.downButt)
			l:Pop()
			l:PushH(LM.GUI.ALIGN_CENTER)
				dlg.removeButt= LM.GUI.Button("Remove",FA_MPlayer.REMOVE)
				l:AddChild(dlg.removeButt)
				dlg.clearButt= LM.GUI.Button("Clear all",FA_MPlayer.CLEAR)
				l:AddChild(dlg.clearButt)
			l:Pop()
		l:Pop()

		l:PushV(LM.GUI.ALIGN_TOP)
			l:PushH(LM.GUI.ALIGN_RIGHT)
				l:AddChild(LM.GUI.StaticText("File:"))
				dlg.name= LM.GUI.TextControl(0,"Placeholder for a longer file",0,LM.GUI.FIELD_TEXT)
				dlg.name:SetValue("")
				l:AddChild(dlg.name)
				l:AddChild(LM.GUI.Button("Browse...",FA_MPlayer.BROWSE),LM.GUI.ALIGN_RIGHT)
			l:Pop()
			l:PushH(LM.GUI.ALIGN_RIGHT)
				l:AddChild(LM.GUI.StaticText("Name for the playlist:"))
				dlg.listName= LM.GUI.TextControl(0,"Placeholder for a name",0,LM.GUI.FIELD_TEXT)
				dlg.listName:SetValue("")
				l:AddChild(dlg.listName)
			l:Pop()
			l:PushH()
				dlg.seqCheck= LM.GUI.CheckBox("Image sequence of:", FA_MPlayer.LENGTH)
				l:AddChild(dlg.seqCheck)
				dlg.seqNum= LM.GUI.TextControl(0,"00000",0,LM.GUI.FIELD_UINT)
				dlg.seqNum:SetWheelInc(1)
				dlg.seqNum:SetValue(1)
				l:AddChild(dlg.seqNum)
				l:AddChild(LM.GUI.StaticText("frames"))
			l:Pop()
			l:PushH()
				l:AddChild(LM.GUI.StaticText("Playing speed:"))
				dlg.speedNum= LM.GUI.TextControl(0,"0000",0,LM.GUI.FIELD_UINT)
				dlg.speedNum:SetWheelInc(1)
				dlg.speedNum:SetValue(FA_MPlayer.defaultSpeed)
				l:AddChild(dlg.speedNum)
				l:AddChild(LM.GUI.StaticText("frames/sec."))
			l:Pop()

			l:AddPadding(10)

			l:PushH(LM.GUI.ALIGN_RIGHT)
				dlg.audioCheck= LM.GUI.CheckBox("Audio file:", FA_MPlayer.AUDIO)
				l:AddChild(dlg.audioCheck)
				dlg.audio= LM.GUI.TextControl(0,"Placehlder for a long",0,LM.GUI.FIELD_TEXT)
				dlg.audio:SetValue("")
				l:AddChild(dlg.audio)
				dlg.audioBrowse= LM.GUI.Button("Browse...",FA_MPlayer.ABROWSE)
				l:AddChild(dlg.audioBrowse,LM.GUI.ALIGN_RIGHT)
			l:Pop()

			l:AddPadding(10)

			l:PushH(LM.GUI.ALIGN_CENTER)
				l:AddChild(LM.GUI.Button("Add to list",FA_MPlayer.ADD))
				dlg.modifyButt= LM.GUI.Button("Modify",FA_MPlayer.MODIFY)
				l:AddChild(dlg.modifyButt)
			l:Pop()
			l:PushH(LM.GUI.ALIGN_CENTER)
				l:AddChild(LM.GUI.Button("Load list file",FA_MPlayer.LOAD))
				dlg.saveButt= LM.GUI.Button("Save to list file",FA_MPlayer.SAVE)
				l:AddChild(dlg.saveButt)
			l:Pop()
		l:Pop()
	l:Pop()

	dlg.tab= {}
	local playList= FA_MPlayer.playList

	for i= 1, table.getn(playList) do
		table.insert(dlg.tab,playList[i])
	end
	dlg.firstCall= true
	return dlg
end

function FA_MPlayerListDialog:UpdateWidgets()

	if (self.firstCall) then
		self.firstCall= false
		self:FillList(0)
	end
	local en= false
	local aen= false

	if (self.seqCheck:Value()) then
		en= true
		aen= self.audioCheck:Value()
	end
	self.seqNum:Enable(en)
	self.speedNum:Enable(en)
	self.audioCheck:Enable(en)
	self.audio:Enable(aen)
	self.audioBrowse:Enable(aen)

	local max= table.getn(self.tab)
	local item= self.list:SelItem()

	self.upButt:Enable(item > 0 and item < max)
	self.downButt:Enable(item < max - 1)

	self.removeButt:Enable(max > 0 and item < max)
	self.clearButt:Enable(max > 0)
	self.modifyButt:Enable(max > 0 and item < max)
	self.saveButt:Enable(max > 0)
end

function FA_MPlayerListDialog:HandleMessage(msg)

	if (msg == LM.GUI.MSG_OK) then
--exit with the OK button
		FA_MPlayer.playList= {}

		for i= 1, table.getn(self.tab) do
			table.insert(FA_MPlayer.playList,self.tab[i])
		end
		FA_MPlayer.dialogOK= true
	elseif (msg == LM.GUI.MSG_CANCEL) then
		FA_MPlayer.dialogOK= false
	elseif (msg == FA_MPlayer.BROWSE) then
		local fileName= LM.GUI.OpenFile("Select the video file to play")

		if (fileName ~= "") then
			self.name:SetValue(fileName)

			local listName= fileName
			local len= string.len(fileName) + 1

			while (len > 1) do
				len= len - 1

				if (string.find(fileName, "[/\\]",len)) then
					listName= string.sub(fileName,len + 1)
					break
				end
			end
			self.listName:SetValue(listName)
			len= self:ScanImportFiles(fileName)

			if (len > 0) then
				self.seqCheck:SetValue(true)
				self.seqNum:SetValue(len)
			else
				self.seqCheck:SetValue(false)
			end
			self:UpdateWidgets()
		end
	elseif (msg == FA_MPlayer.ABROWSE) then
		local fileName= LM.GUI.OpenFile("Select the audio file to play")

		if (fileName ~= "") then
			self.audio:SetValue(fileName)
		end
	elseif (msg == FA_MPlayer.LIST) then
		self:ListSelected(self.list:SelItem())
	elseif (msg == FA_MPlayer.LENGTH) then
		self:UpdateWidgets()
	elseif (msg == FA_MPlayer.AUDIO) then
		self:UpdateWidgets()
	elseif (msg == FA_MPlayer.ADD) then
		self:SetItem(false)
	elseif (msg == FA_MPlayer.MODIFY) then
		self:SetItem(true)
	elseif (msg == FA_MPlayer.LOAD) then
		self:LoadList()
	elseif (msg == FA_MPlayer.SAVE) then
		self:SaveList()
	elseif (msg == FA_MPlayer.UP) then
		local cur= self.list:SelItem() + 1

		if (cur > 1) then
			table.insert(self.tab, cur - 1, self.tab[cur])
			table.remove(self.tab, cur + 1)
			self:FillList(cur - 2)
		end

	elseif (msg == FA_MPlayer.DOWN) then
		local cur= self.list:SelItem() + 1

		if (cur < table.getn(self.tab)) then
			table.insert(self.tab, cur + 2, self.tab[cur])
			table.remove(self.tab, cur)
			self:FillList(cur)
		end

	elseif (msg == FA_MPlayer.REMOVE) then
		local cur= self.list:SelItem() + 1
		table.remove(self.tab, cur)
		self:FillList(cur - 1)
	elseif (msg == FA_MPlayer.CLEAR) then
		self.tab= {}
		self:FillList(0)
	end
end

function FA_MPlayerListDialog:FillList(cursor)

--fill the name list

	local list= self.list
	local len= list:CountItems()
	local tlen= table.getn(self.tab)

	for i= 0, tlen - 1 do
		local name= self.tab[i + 1][3]

		if (i < len) then

			if (list:GetItem(i) ~= name) then
				list:SetItemLabel(i,name)
			end
		else
			list:AddItem(name)
		end
	end

	for i= len - 1, tlen, -1 do
		list:RemoveItem(i)
	end

	if (tlen > 0) then
		list:AddItem(" ")
--set cursor
		local selKey= list:GetItem(cursor)
		list:SetItemLabel(cursor,"*")
		list:SetSelItem("*")
		list:SetItemLabel(cursor,selKey)

		self:ListSelected(cursor)
	end
end

function FA_MPlayerListDialog:ListSelected(cursor)

--update the widgets by the selected list item

	if (cursor < table.getn(self.tab)) then
		local tab= self.tab[cursor + 1]
		local mode= tab[1]

		self.name:SetValue(tab[2])
		self.listName:SetValue(tab[3])

		if (mode == 0) then
			self.seqCheck:SetValue(false)
		else
			self.seqCheck:SetValue(true)
			self.seqNum:SetValue(tab[4])
			self.speedNum:SetValue(tab[5])

			if (mode == 2) then
				self.audio:SetValue(tab[6])
				self.audioCheck:SetValue(true)
			else
				self.audioCheck:SetValue(false)
			end
		end
	end
	self:UpdateWidgets()
end

function FA_MPlayerListDialog:LoadList()

	local fileName= LM.GUI.OpenFile("Select list file to load")

	if (fileName == "") then
		return
	end
	local f= io.open(fileName,"r")

	if (f == nil) then
		LM.GUI.Alert(LM.GUI.ALERT_INFO,"Cannot open file.")
		return
	end
	f:close()

	sTab= {}
	local tab= {}
	local tLen= 0
	local tType= 0

--the item successfully loaded when tLen < 0

	for line in io.lines(fileName) do
		local c= string.sub(line,1,1)

		if (line ~= "") then

			if (c == "=") then
	--initial item in the table
				if (tLen < 0) then
					table.insert(sTab, tab)
				end
				tType= tonumber(string.sub(line,2))
				tab= { tType }
				tLen= 1

			elseif (c == ">" and tLen > 0) then
	--other values in the table
				tLen= tLen + 1

				if (tLen < 4) then
					table.insert(tab,string.sub(line,2))

					if (tLen == 3 and tType == 0) then
						tLen= -1
					end

				elseif ((tType == 1 or tType == 2) and tLen < 6) then
		--numeric parameters
					table.insert(tab,tonumber(string.sub(line,2)))

					if (tLen == 5 and tType == 1) then
						tLen= -1
					end

				elseif (tType == 2 and tLen == 5) then
		--audio file
					table.insert(tab,string.sub(line,2))
					tLen= -1
				end
			elseif (c ~= ";") then
		--not a remark, so wrong
				LM.GUI.Alert(LM.GUI.ALERT_INFO,"Bad file format.")
				return
			end
		end
	end

	if (tLen < 0) then
		table.insert(sTab, tab)
	end
	self.tab= sTab
	self:FillList(0)
end

function FA_MPlayerListDialog:SaveList()

	local fileName= LM.GUI.OpenFile("Select list file to save")

	if (fileName == "") then
		return
	end

	local f= io.open(fileName,"w")

	if (f == nil) then
		LM.GUI.Alert(LM.GUI.ALERT_INFO,"Cannot open file for writing.")
		return
	end
	f:write("; --- play list for fa_mplayer.lua (moho) - do not edit by hand ---\n")

	for i= 1, table.getn(self.tab) do
		local tab= self.tab[i]
		f:write("="..tostring(tab[1]).."\n")

		for j= 2, table.getn(tab) do
			f:write(">"..tostring(tab[j]).."\n")
		end
		f:write("\n")
	end
	f:close()
end

function FA_MPlayerListDialog:SetItem(modify)

	local name= self.name:Value()
	local f= io.open(name)

	if (f) then
		f:close()
	else
		LM.GUI.Alert(LM.GUI.ALERT_INFO,"Please specify a valid filename first.")
		return
	end
	local listName= self.listName:Value()

	if (listName == "") then
		LM.GUI.Alert(LM.GUI.ALERT_INFO,"Please specify a name for the playlist.")
		return
	end
	local len= 0

	if (self.seqCheck:Value()) then
		len= self.seqNum:IntValue()
	end
	local tab

	if (len > 0) then
		local fps= self.speedNum:IntValue()
		local audio= ""

		if (self.audioCheck:Value()) then
			audio= self.audio:Value()
		end

		if (audio == "") then
			tab= { 1, name, listName, len, fps }
		else
			tab= { 2, name, listName, len, fps, audio }
		end
	else
		tab= { 0, name, listName }
	end
	local pos= self.list:SelItem()
	len= table.getn(self.tab)

	if (len > 0 and len >= pos) then
		table.insert(self.tab, pos + 1, tab)

		if (modify) then
			table.remove(self.tab, pos + 2)
		else
			pos= pos + 1
		end
	else
		if (modify) then
			return
		end
		table.insert(self.tab, tab)
		pos= len + 1
	end
	self:FillList(pos)
end

function FA_MPlayerListDialog:ScanImportFiles(fName)

--scan the number of files in the image sequence

	local ext= string.lower(string.sub(fName,-4))

	if (ext ~= ".jpg" and ext ~= ".png" and ext ~= ".tga") then
		return 0
	end
	local lastNumber= string.len(fName) + 1

	while (lastNumber > 1) do
		lastNumber= lastNumber - 1

		if (string.find(fName,"%d",lastNumber)) then
			lastNumber= lastNumber + 1
			break
		end

		if (string.find(fName,"[/\\]",lastNumber)) then
			lastNumber= 0
			break
		end
	end
	local lastNEnd= lastNumber

	while (lastNumber > 1) do
		lastNumber= lastNumber - 1

		if (string.find(fName,"%d",lastNumber) ~= lastNumber) then
			lastNumber= lastNumber + 1
			break
		end
	end
	local cnt= 1

	if (lastNEnd > lastNumber) then
--numeric field exists
		local n= tonumber(string.sub(fName,lastNumber,lastNEnd - 1))
		local digits= lastNEnd - lastNumber
		local pre= string.sub(fName,1,lastNumber-1)
		local post= string.sub(fName,lastNEnd)

		while true do
			n= n + 1
			local f= io.open(pre..string.format("%0"..digits.."d",n)..post, "r")

			if (f == nil) then
				break
			end
			f:close()
			cnt= cnt + 1
		end
	end
	return cnt
end

-- **************************************************
-- options dialog window
-- **************************************************

local FA_MPlayerOptionsDialog = {}

function FA_MPlayerOptionsDialog:new()
	local dlg = LM.GUI.SimpleDialog("MPlayer Options", FA_MPlayerOptionsDialog)

	if (not dlg) then
		return nil
	end
	local l = dlg:GetLayout()

	l:PushH(LM.GUI.ALIGN_RIGHT)
		l:AddChild(LM.GUI.StaticText("MPlayer command:"))
		dlg.cmdText= LM.GUI.TextControl(0,"Placeholder for a very long filename to invoke mplayer command",0,LM.GUI.FIELD_TEXT)
		dlg.cmdText:SetValue("")
		l:AddChild(dlg.cmdText)
	l:Pop()
	dlg.exeCheck= LM.GUI.CheckBox("Use the os.execute() LUA function instead of io.popen()",FA_MPlayer.OSEXE)
	l:AddChild(dlg.exeCheck, LM.GUI.ALIGN_LEFT)
	l:AddChild(LM.GUI.StaticText("(You must create a fifo device file to use the os.execute() mode)"))

	l:PushH(LM.GUI.ALIGN_RIGHT)
		l:AddChild(LM.GUI.StaticText("Fifo file for os.execute() :"))
		dlg.fifoText= LM.GUI.TextControl(0,"Placeholder for a very long filename")
		dlg.fifoText:SetValue("")
		l:AddChild(dlg.fifoText)
	l:Pop()

	l:PushH(LM.GUI.ALIGN_RIGHT)
		l:AddChild(LM.GUI.StaticText("Invoking style for os.execute() :"))
		dlg.exeText= LM.GUI.TextControl(0,"Placeholder for a very long filename")
		dlg.exeText:SetValue("")
		l:AddChild(dlg.exeText)
	l:Pop()

	l:AddChild(LM.GUI.Button("Reset", FA_MPlayer.RESET))
	return dlg
end

function FA_MPlayerOptionsDialog:UpdateWidgets()
	local en= FA_MPlayer.MP_ExeMode

	self.cmdText:SetValue(FA_MPlayer.MP_Invoke)
	self.exeCheck:SetValue(en)
	self.fifoText:SetValue(FA_MPlayer.MP_Fifo)
	self.exeText:SetValue(FA_MPlayer.MP_ExeFormat)

	self.fifoText:Enable(en)
	self.exeText:Enable(en)
end

function FA_MPlayerOptionsDialog:HandleMessage(msg)

	if (msg == LM.GUI.MSG_OK) then
--exit with the OK button
		FA_MPlayer.MP_Invoke = self.cmdText:Value()
		FA_MPlayer.MP_ExeMode = self.exeCheck:Value()
		FA_MPlayer.MP_Fifo = self.fifoText:Value()
		FA_MPlayer.MP_ExeFormat = self.exeText:Value()

	elseif (msg == FA_MPlayer.OSEXE) then
		local en= self.exeCheck:Value()
		self.fifoText:Enable(en)
		self.exeText:Enable(en)
	elseif (msg == FA_MPlayer.RESET) then
		FA_MPlayer:SetInvoke()
		self:UpdateWidgets()
	end
end

-- **************************************************
-- MPlayer control routines
-- **************************************************

function FA_MPlayer:QuitPlayer()

	if (self.MP_On) then
		self.MP_File:write(self.MP_Quit)
		self.MP_File:close()
		self.MP_On= false
	end
	self.MP_Playing= false
end

function FA_MPlayer:SetOSD(mode)

--mode= 0: forced clear/set, 1: clear if on, 2: set if on

	if (self.MP_On and self.MP_OSD_Enabled) then
		local osd= self.MP_OSDMode

		if (mode ~= 0) then

			if (not osd) then
				return
			end

			if (mode == 1) then
				osd= false 
			end
		end

		if (osd) then

			if (self.MP_Playing) then
				self.MP_File:write(self.MP_OSD_On)
			else
				self.MP_File:write(self.MP_OSD_On_Pause)
			end
		else
			if (self.MP_Playing) then
				self.MP_File:write(self.MP_OSD_Off)
			else
				self.MP_File:write(self.MP_OSD_Off_Pause)
			end
		end
		self.MP_File:flush()
	end
end

function FA_MPlayer:CheckFile(name)
	local f= io.open(name)

	if (f) then
		f:close()
		return true
	end
	return false
end

function FA_MPlayer.ExeReplace(key)

	if (key == "cmd") then
		return FA_MPlayer.MP_FullInvoke
	elseif (key == "fifo") then
		return FA_MPlayer.MP_Fifo
	end
end

function FA_MPlayer:RestartPlayer()
	self:QuitPlayer()
	local hasFile= false
	local invoke= self.MP_Invoke

	for i= 1, table.getn(self.playList) do
		local itemType= self.playList[i][1]
		local name= self.playList[i][2]

		if (itemType == 0) then
	--movie file
			if (self:CheckFile(name)) then
				invoke= invoke..string.format(self.MP_PlayFile,name)
				hasFile= true
			end
		else
	--image sequence
			if (self:CheckFile(name)) then
				local ext= string.lower(string.sub(name,-3))

				if (ext ~= "png" and ext ~= "tga") then
					ext= "jpg"
				end
				local lastNumber= string.len(name) + 1

				while (lastNumber > 1) do
					lastNumber= lastNumber - 1

					if (string.find(name,"%d",lastNumber)) then
						lastNumber= lastNumber + 1
						break
					end

					if (string.find(name,"[/\\]",lastNumber)) then
						lastNumber= 0
						break
					end
				end
				local lastNEnd= lastNumber

				while (lastNumber > 1) do
					lastNumber= lastNumber - 1

					if (string.find(name,"%d",lastNumber) ~= lastNumber) then
						lastNumber= lastNumber + 1
						break
					end
				end

		--creating a filename list (long strings could be problems!)

				if (lastNEnd > lastNumber) then
					local n= tonumber(string.sub(name,lastNumber,lastNEnd - 1))
					local digits= lastNEnd - lastNumber
					local pre= string.sub(name,1,lastNumber - 1)
					local post= string.sub(name,lastNEnd)

					for j= 1, self.playList[i][4] do
						name= name..","..pre..string.format("%0"..digits.."d",n)..post
						n= n + 1
					end
				end
				invoke= invoke..string.format(self.MP_PlaySeq,name,self.playList[i][5],ext)

				if (itemType == 2) then
		--with sound
					name= self.playList[i][6]

					if (self:CheckFile(name)) then
						invoke= invoke..string.format(self.MP_PlaySound,name)
					end
				end
				hasFile= true
			end
		end
	end

	if (hasFile) then
		self.MP_FullInvoke= invoke

		if (self.MP_ExeMode) then
			invoke= string.gsub( self.MP_ExeFormat, "%$(%w+)", self.ExeReplace )
			os.execute(invoke)
			self.MP_File= io.open(self.MP_Fifo,"a+")
		else
			self.MP_File= io.popen(invoke,"w")
		end

		if (self.MP_File) then
			self.MP_File:write(self.MP_Init)
			self.MP_File:flush()
			self.MP_On= true
			self:SetOSD(2)
			return true
		else
			LM.GUI.Alert(LM.GUI.ALERT_INFO,"Cannot invoke MPlayer. Please check the MPlayer Options.")
		end
	end
	return false
end

function FA_MPlayer:CheckOS()
	local env= os.getenv("OSTYPE")

	if (env) then

		if (string.lower(string.sub(env, 1, 5)) == "linux") then
			return "linux"
		end
	end
	env= os.getenv("WINDIR")

	if (env) then
		return "windows"
	end
	return "mac"
end

function FA_MPlayer:SetInvoke()

	local os= self.CheckOS()

	if (os == "windows") then
		self.MP_Invoke = self.MP_Windows_Invoke
		self.MP_ExeFormat = "start $cmd <$fifo"
		self.MP_Fifo = "C:\\moho_fifo"
		self.MP_ExeMode = false
	else
		self.MP_ExeFormat = "$cmd <$fifo &"
		self.MP_Fifo = "/dev/moho_fifo"

		if (os == "linux") then
			self.MP_Invoke = self.MP_Linux_Invoke
			self.MP_ExeMode = false
		else
			self.MP_Invoke = self.MP_Mac_Invoke
			self.MP_ExeMode = true
		end
	end
end

-- **************************************************
-- dialog window
-- **************************************************

local FA_MPlayerDialog = {}

function FA_MPlayerDialog:new()
	local dlg = LM.GUI.SimpleDialog("MPlayer control", FA_MPlayerDialog)

	if (not dlg) then
		return nil
	end

	local l= dlg:GetLayout()

	dlg.osdCheck= nil

	l:PushH()
		if (FA_MPlayer.MP_OSD_Enabled) then
			dlg.osdCheck= LM.GUI.CheckBox("On-Screen Display", FA_MPlayer.OSD)
			l:AddChild(dlg.osdCheck)
		end
		l:AddChild(LM.GUI.Button("Set playlist...", FA_MPlayer.PLAYLIST))
		l:AddChild(LM.GUI.Button("MPlayer options", FA_MPlayer.OPTIONS))
	l:Pop()

	l:PushH()
		l:AddChild(LM.GUI.Button("|<< Restart", FA_MPlayer.RESTART))
		l:AddChild(LM.GUI.StaticText("Seek:"))
		dlg.seek= LM.GUI.AngleWidget(FA_MPlayer.SEEK)
		dlg.lastSeek= 0.0
		dlg.lastClock= os.clock()
		dlg.seek:SetValue(0.0)
		l:AddChild(dlg.seek)
		dlg.mark= LM.GUI.RadioButton("",FA_MPlayer.MARK)
		l:AddChild(dlg.mark)
		l:AddChild(LM.GUI.Button("Play > / Pause ||", FA_MPlayer.PAUSE))
		l:AddChild(LM.GUI.Button("||< Step",FA_MPlayer.STEPBACK))
		l:AddChild(LM.GUI.Button("Step >||",FA_MPlayer.STEP))
		l:AddChild(LM.GUI.Button("Close", LM.GUI.MSG_CANCEL))
	l:Pop()

	return dlg
end

function FA_MPlayerDialog:UpdateWidgets()

	if (self.osdCheck) then
		self.osdCheck:SetValue(FA_MPlayer.MP_OSDMode)
	end
	self.mark:SetValue(FA_MPlayer.MP_Playing)
end

function FA_MPlayerDialog:OnOK()
	FA_MPlayer:QuitPlayer()
end

function FA_MPlayerDialog:HandleMessage(msg)

	if (msg == LM.GUI.MSG_OK or msg == LM.GUI.MSG_CANCEL) then
		FA_MPlayer:QuitPlayer()
	elseif (msg == FA_MPlayer.PLAYLIST) then
		self:PlayList()
	elseif (msg == FA_MPlayer.OPTIONS) then
		local dlg= FA_MPlayerOptionsDialog:new()

		if (dlg ~= nil) then
			local on= FA_MPlayer.MP_On

			if (on) then
				FA_MPlayer:QuitPlayer()
				self:UpdateWidgets()
			end
			dlg:DoModal()

			if (on) then
				FA_MPlayer:RestartPlayer()
				self:UpdateWidgets()
			end
		end

	elseif (msg == FA_MPlayer.OSD) then
		FA_MPlayer.MP_OSDMode= self.osdCheck:Value()
		FA_MPlayer:SetOSD(0)
	elseif (msg == FA_MPlayer.RESTART) then
		FA_MPlayer:RestartPlayer()
		self:UpdateWidgets()
	else
-- control commands

		if (not FA_MPlayer.MP_On) then

			if (msg == FA_MPlayer.MARK) then
				self.mark:SetValue(false)
			end
			return
		end
		local f= FA_MPlayer.MP_File

		if (msg == FA_MPlayer.SEEK) then
--seek "jog wheel"
			local angle= self.seek:Value() - self.lastSeek
			local seek

			if (angle > math.pi) then
				seek= LM.Round(-10 * (angle - 2 * math.pi))
			elseif (angle < -math.pi) then
				seek= LM.Round(-10 * (angle + 2 * math.pi))
			else
				seek= LM.Round(-10 * angle)
			end

			if (seek > 1 or seek < -1) then
	--prevent from delay effects
				local clock= os.clock()

				if (clock - self.lastClock ~= 0) then
					self.lastClock= clock
					self.lastSeek= self.lastSeek + angle

					if (FA_MPlayer.MP_Playing) then
						f:write(string.format(FA_MPlayer.MP_Seek,seek))
					else
						f:write(string.format(FA_MPlayer.MP_Seek_Pause,seek))
					end
					f:flush()
				end
			end

		elseif (msg == FA_MPlayer.PAUSE or msg == FA_MPlayer.MARK) then
			FA_MPlayer.MP_Playing= not FA_MPlayer.MP_Playing

			if (FA_MPlayer.MP_Playing) then
				f:write(FA_MPlayer.MP_Start)
			else
				f:write(FA_MPlayer.MP_Stop)
			end
			f:flush()
			self:UpdateWidgets()

		elseif (msg == FA_MPlayer.STEP) then
			f:write(FA_MPlayer.MP_Step)
			f:flush()
			FA_MPlayer.MP_Playing= false
			self:UpdateWidgets()
		elseif (msg == FA_MPlayer.STEPBACK) then
			f:write(FA_MPlayer.MP_Step_Back)
			f:flush()
			FA_MPlayer.MP_Playing= false
			self:UpdateWidgets()
		end
	end
end

function FA_MPlayerDialog:PlayList()

	local dlg= FA_MPlayerListDialog:new()
	FA_MPlayer.dialogOK= false

	if (dlg ~= nil) then
		dlg:DoModal()

		if (FA_MPlayer.dialogOK) then
			FA_MPlayer:RestartPlayer()
		end
	end
end

-- **************************************************
-- Tool options - create and respond to tool's UI
-- **************************************************

FA_MPlayer.RESTART = MOHO.MSG_BASE
FA_MPlayer.SEEK = MOHO.MSG_BASE + 1
FA_MPlayer.PAUSE = MOHO.MSG_BASE + 2
FA_MPlayer.MARK = MOHO.MSG_BASE + 3
FA_MPlayer.STEP = MOHO.MSG_BASE + 4
FA_MPlayer.STEPBACK = MOHO.MSG_BASE + 5
FA_MPlayer.OSD = MOHO.MSG_BASE + 6
FA_MPlayer.PLAYLIST = MOHO.MSG_BASE + 7
FA_MPlayer.OPTIONS = MOHO.MSG_BASE + 8

FA_MPlayer.LIST = MOHO.MSG_BASE + 10
FA_MPlayer.BROWSE = MOHO.MSG_BASE + 11
FA_MPlayer.LENGTH = MOHO.MSG_BASE + 12
FA_MPlayer.SPEED = MOHO.MSG_BASE + 13
FA_MPlayer.AUDIO = MOHO.MSG_BASE + 14
FA_MPlayer.ABROWSE = MOHO.MSG_BASE + 15
FA_MPlayer.ADD = MOHO.MSG_BASE + 16
FA_MPlayer.MODIFY = MOHO.MSG_BASE + 17
FA_MPlayer.UP = MOHO.MSG_BASE + 18
FA_MPlayer.DOWN = MOHO.MSG_BASE + 19
FA_MPlayer.REMOVE = MOHO.MSG_BASE + 20
FA_MPlayer.CLEAR = MOHO.MSG_BASE + 21
FA_MPlayer.LOAD = MOHO.MSG_BASE + 22
FA_MPlayer.SAVE = MOHO.MSG_BASE + 23
FA_MPlayer.OSEXE = MOHO.MSG_BASE + 24
FA_MPlayer.RESET = MOHO.MSG_BASE + 25

FA_MPlayer.defaultSpeed= 25
FA_MPlayer.dialogOK= false

function FA_MPlayer:Run(moho)

	if (self.MP_Invoke == "") then
		self:SetInvoke()
	end
	self.defaultSpeed= moho.document:Fps()

	local dlg= FA_MPlayerDialog:new()

	if (dlg ~= nil) then
		self:RestartPlayer()
		dlg:DoModeless()
	end
end

function FA_MPlayer:DoLayout(moho)
	self:Run(moho)
end
